import math

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import ModuleCustomSD


class LoConModule(ModuleCustomSD):
    """
    modifed from kohya-ss/sd-scripts/networks/lora:LoRAModule
    """

    def __init__(
        self,
        lora_name,
        org_module: nn.Module,
        multiplier=1.0,
        lora_dim=4,
        alpha=1,
        dropout=0.0,
        rank_dropout=0.0,
        module_dropout=0.0,
        use_tucker=False,
        use_scalar=False,
        rank_dropout_scale=False,
        **kwargs,
    ):
        """if alpha == 0 or None, alpha is rank (no scaling)."""
        super().__init__()
        self.lora_name = lora_name
        self.lora_dim = lora_dim
        self.tucker = False

        if isinstance(org_module, nn.Conv2d):
            self.isconv = True
            # For general LoCon
            in_dim = org_module.in_channels
            k_size = org_module.kernel_size
            stride = org_module.stride
            padding = org_module.padding
            out_dim = org_module.out_channels
            self.down_op = F.conv2d
            self.up_op = F.conv2d
            if use_tucker and k_size != (1, 1):
                self.lora_down = nn.Conv2d(in_dim, lora_dim, (1, 1), bias=False)
                self.lora_mid = nn.Conv2d(
                    lora_dim, lora_dim, k_size, stride, padding, bias=False
                )
                self.tucker = True
            else:
                self.lora_down = nn.Conv2d(
                    in_dim, lora_dim, k_size, stride, padding, bias=False
                )
            self.lora_up = nn.Conv2d(lora_dim, out_dim, (1, 1), bias=False)
        elif isinstance(org_module, nn.Linear):
            self.isconv = False
            self.down_op = F.linear
            self.up_op = F.linear
            in_dim = org_module.in_features
            out_dim = org_module.out_features
            self.lora_down = nn.Linear(in_dim, lora_dim, bias=False)
            self.lora_up = nn.Linear(lora_dim, out_dim, bias=False)
        else:
            raise NotImplementedError
        self.shape = org_module.weight.shape

        if dropout:
            self.dropout = nn.Dropout(dropout)
        else:
            self.dropout = nn.Identity()
        self.rank_dropout = rank_dropout
        self.rank_dropout_scale = rank_dropout_scale
        self.module_dropout = module_dropout

        if type(alpha) == torch.Tensor:
            alpha = alpha.detach().float().numpy()  # without casting, bf16 causes error
        alpha = lora_dim if alpha is None or alpha == 0 else alpha
        self.scale = alpha / self.lora_dim
        self.register_buffer("alpha", torch.tensor(alpha))  # 定数として扱える

        if use_scalar:
            self.scalar = nn.Parameter(torch.tensor(0.0))
        else:
            self.scalar = torch.tensor(1.0)
        # same as microsoft's
        torch.nn.init.kaiming_uniform_(self.lora_down.weight, a=math.sqrt(5))
        if use_scalar:
            torch.nn.init.kaiming_uniform_(self.lora_up.weight, a=math.sqrt(5))
        else:
            torch.nn.init.constant_(self.lora_up.weight, 0)
        if self.tucker:
            torch.nn.init.kaiming_uniform_(self.lora_mid.weight, a=math.sqrt(5))

        self.multiplier = multiplier
        self.org_module = [org_module]
        self.org_forward = self.org_module[0].forward

    def load_weight_hook(self, module: nn.Module, incompatible_keys):
        missing_keys = incompatible_keys.missing_keys
        for key in missing_keys:
            if "scalar" in key:
                del missing_keys[missing_keys.index(key)]
        self.scalar = nn.Parameter(torch.ones_like(self.scalar))

    def apply_to(self, is_hypernet=False, **kwargs):
        self.org_forward = self.org_module[0].forward
        if is_hypernet:
            self.org_module[0].forward = self.hypernet_forward
        else:
            self.org_module[0].forward = self.forward

    def restore(self):
        self.org_module[0].forward = self.org_forward

    def merge_to(self, multiplier=1.0):
        weight = self.make_weight() * self.scale * multiplier
        self.org_module[0].weight.data.add_(weight)

    def make_weight(self, device=None):
        wa = self.lora_up.weight.to(device)
        wb = self.lora_down.weight.to(device)
        return (wa.view(wa.size(0), -1) @ wb.view(wb.size(0), -1)).view(
            self.shape
        ) * self.scalar

    def custom_state_dict(self):
        destination = {}
        destination["alpha"] = self.alpha
        destination["lora_up.weight"] = self.lora_up.weight * self.scalar
        destination["lora_down.weight"] = self.lora_down.weight
        if self.tucker:
            destination["lora_mid.weight"] = self.lora_mid.weight
        return destination

    @torch.no_grad()
    def apply_max_norm(self, max_norm, device=None):
        orig_norm = self.make_weight(device).norm() * self.scale
        norm = torch.clamp(orig_norm, max_norm / 2)
        desired = torch.clamp(norm, max=max_norm)
        ratio = desired.cpu() / norm.cpu()

        scaled = ratio.item() != 1.0
        if scaled:
            self.scalar *= ratio

        return scaled, orig_norm * ratio

    def update_weights(self, down, up, idx):
        self.down, self.up = self.make_lightweight(down.squeeze(1), up.squeeze(1), idx)

    def make_lightweight(self, down, up, seed=None, down_aux=None, up_aux=None):
        if down.dim() == 3:
            down = down.reshape(down.size(0), self.lora_dim, -1)
            up = up.reshape(up.size(0), -1, self.lora_dim)
        else:
            down = down.reshape(self.lora_dim, -1)
            up = up.reshape(-1, self.lora_dim)
        # print(up.shape)
        if seed is None:
            assert down_aux is not None and up_aux is not None
        else:
            rng_state = torch.get_rng_state()
            torch.manual_seed(seed)
            if down_aux is None or up_aux is None:
                down_aux = torch.empty(
                    down.size(down.dim() - 1),
                    self.lora_down.weight.size(1),
                    device=down.device,
                )
                up_aux = torch.empty(
                    self.lora_up.weight.size(0), up.size(up.dim() - 2), device=up.device
                )
                nn.init.orthogonal_(down_aux)
                nn.init.orthogonal_(up_aux)
                # print(up_aux.shape)
            torch.set_rng_state(rng_state)
        if down.dim() == 3 and down.size(0) == 1:
            down = down.squeeze(0)
        if up.dim() == 3 and up.size(0) == 1:
            up = up.squeeze(0)
        down = down + 1  # avoid zero grad or slow training, give it a constant
        return (down @ down_aux), (up_aux @ up)

    def apply_lightweight(self, down, up, seed=None, down_aux=None, up_aux=None):
        down_weight, up_weight = self.make_lightweight(down, up, seed, down_aux, up_aux)
        self.lora_down.weight.data = down_weight
        self.lora_up.weight.data = up_weight
        return down_weight, up_weight

    def hypernet_forward(self, x):
        if self.module_dropout and self.training:
            if torch.rand(1) < self.module_dropout:
                return self.org_forward(x)
        scale = self.scale * self.multiplier

        down_weight = self.down
        up_weight = self.up

        x_batch = None
        if down_weight.dim() == 3:
            if x.size(0) != down_weight.size(0):
                assert (
                    self.isconv == False
                ), "Convolutional hypernet with batch size mismatch is not supported"
                x_batch = x.size(0)
                x = x.view(down_weight.size(0), -1, *x.shape[1:])

            if self.isconv:
                mid = torch.einsum("ijk, ik... -> ij...", down_weight, x)
            else:
                mid = torch.einsum("ijk, i...k -> i...j", down_weight, x)
        else:
            if self.isconv:
                weight = down_weight.unsqueeze(-1).unsqueeze(-1)
            else:
                weight = down_weight
            mid = self.down_op(x, weight)

        if self.rank_dropout and self.training:
            drop = (
                torch.rand(self.lora_dim, device=mid.device) < self.rank_dropout
            ).to(mid.dtype)
            if self.rank_dropout_scale:
                drop /= drop.mean()
            if (dims := len(x.shape)) == 4:
                drop = drop.view(1, -1, 1, 1)
            else:
                drop = drop.view(*[1] * (dims - 1), -1)
            mid = mid * drop

        if up_weight.dim() == 3:
            mid_batch = None
            if mid.size(0) != up_weight.size(0):
                assert (
                    self.isconv == False
                ), "Convolutional hypernet with batch size mismatch is not supported"
                mid_batch = mid.size(0)
                mid = mid.view(up_weight.size(0), -1, *mid.shape[1:])

            if self.isconv:
                up = torch.einsum("ijk, ik... -> ij...", up_weight, mid)
            else:
                up = torch.einsum("ijk, i...k -> i...j", up_weight, mid)

            if mid_batch is not None:
                up = up.view(mid_batch, *up.shape[2:])
        else:
            if self.isconv:
                weight = up_weight.unsqueeze(-1).unsqueeze(-1)
            else:
                weight = up_weight
            up = self.up_op(mid, weight)

        if x_batch is not None:
            up = up.view(x_batch, *up.shape[2:])

        org_out = self.org_forward(x)
        # print(x.shape, org_out.shape, up.shape)
        return org_out + self.dropout(up) * scale

    def forward(self, x):
        if self.module_dropout and self.training:
            if torch.rand(1) < self.module_dropout:
                return self.org_forward(x)
        scale = self.scale * self.multiplier
        if self.tucker:
            mid = self.lora_mid(self.lora_down(x))
        else:
            mid = self.lora_down(x)

        if self.rank_dropout and self.training:
            drop = (
                torch.rand(self.lora_dim, device=mid.device) > self.rank_dropout
            ).to(mid.dtype)
            if self.rank_dropout_scale:
                drop /= drop.mean()
            if (dims := len(x.shape)) == 4:
                drop = drop.view(1, -1, 1, 1)
            else:
                drop = drop.view(*[1] * (dims - 1), -1)
            mid = mid * drop

        return self.org_forward(x) + self.dropout(
            self.lora_up(mid) * self.scalar * scale
        )



import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import ModuleCustomSD


class IA3Module(ModuleCustomSD):
    """
    Hadamard product Implementaion for Low Rank Adaptation
    """

    def __init__(
        self,
        lora_name,
        org_module: nn.Module,
        multiplier=1.0,
        train_on_input=False,
        **kwargs
    ):
        """if alpha == 0 or None, alpha is rank (no scaling)."""
        super().__init__()
        self.lora_name = lora_name
        self.cp = False

        self.shape = org_module.weight.shape
        if org_module.__class__.__name__ == "Conv2d":
            in_dim = org_module.in_channels
            out_dim = org_module.out_channels
            if train_on_input:
                train_dim = in_dim
            else:
                train_dim = out_dim
            self.weight = nn.Parameter(torch.empty(1, train_dim, 1, 1))
        else:
            in_dim = org_module.in_features
            out_dim = org_module.out_features
            if train_on_input:
                train_dim = in_dim
            else:
                train_dim = out_dim

            self.weight = nn.Parameter(torch.empty(train_dim))

        # Need more experiences on init method
        torch.nn.init.constant_(self.weight, 0)

        self.multiplier = multiplier
        self.org_forward = None
        self.org_module = [org_module]  # remove in applying
        self.grad_ckpt = False
        self.train_input = train_on_input
        self.register_buffer("on_input", torch.tensor(int(train_on_input)))

    def apply_to(self):
        self.org_forward = self.org_module[0].forward
        self.org_module[0].forward = self.forward

    @torch.enable_grad()
    def forward(self, x):
        if self.train_input:
            x = x * (1 + self.weight * self.multiplier)
        out = self.org_forward(x)
        dtype = out.dtype
        if not self.train_input:
            out = out * (1 + self.weight * self.multiplier)
            out = out.to(dtype)
        return out



from setuptools import setup, find_packages


setup(
    name="lycoris_lora",
    packages=find_packages(),
    version="2.1.0.post1",
    url="https://github.com/KohakuBlueleaf/LyCORIS",
    description="Lora beYond Conventional methods, Other Rank adaptation Implementations for Stable diffusion",
    author="Shih-Ying Yeh(KohakuBlueLeaf), Yu-Guan Hsieh, Zhidong Gao",
    author_email="apolloyeh0123@gmail.com",
    zip_safe=False,
    install_requires=["torch", "safetensors", "diffusers", "transformers"],
)



from functools import cache
from math import log2, ceil

import torch
import torch.nn as nn
import torch.nn.functional as F

from einops import rearrange

from .base import ModuleCustomSD
from lycoris.logging import logger


@cache
def log_butterfly_factorize(dim, factor, result):
    logger.info(
        f"Use BOFT({int(log2(result[1]))}, {result[0]//2})"
        f" (equivalent to factor={result[0]}) "
        f"for {dim=} and {factor=}"
    )


def butterfly_factor(dimension: int, factor: int = -1) -> tuple[int, int]:
    """
    m = 2k
    n = 2**p
    m*n = dim
    """

    # Find the first solution and check if it is even doable
    m = n = 0
    while m <= factor:
        m += 2
        while dimension % m != 0 and m < dimension:
            m += 2
        if m > factor:
            break
        if sum(int(i) for i in f"{dimension//m:b}") == 1:
            n = dimension // m

    if n == 0:
        raise ValueError(
            f"It is impossible to decompose {dimension} with factor {factor} under BOFT constrains."
        )

    log_butterfly_factorize(dimension, factor, (dimension // n, n))
    return dimension // n, n


class ButterflyOFTModule(ModuleCustomSD):
    """
    modifed from kohya-ss/sd-scripts/networks/lora:LoRAModule
    """

    def __init__(
        self,
        lora_name,
        org_module: nn.Module,
        multiplier=1.0,
        lora_dim=4,
        alpha=1,
        dropout=0.0,
        rank_dropout=0.0,
        module_dropout=0.0,
        use_tucker=False,
        use_scalar=False,
        rank_dropout_scale=False,
        constrain=0,
        rescaled=False,
        **kwargs,
    ):
        """if alpha == 0 or None, alpha is rank (no scaling)."""
        super().__init__()
        self.lora_name = lora_name

        if isinstance(org_module, nn.Linear):
            self.op = F.linear
            self.dim = org_module.out_features
            self.kw_dict = {}
        elif isinstance(org_module, nn.Conv2d):
            self.op = F.conv2d
            self.dim = org_module.out_channels
            self.kw_dict = {
                "stride": org_module.stride,
                "padding": org_module.padding,
                "dilation": org_module.dilation,
                "groups": org_module.groups,
            }
        else:
            raise NotImplementedError

        out_dim = org_module.weight.shape[0]
        b, m_exp = butterfly_factor(out_dim, lora_dim)
        self.block_size = b
        self.block_num = m_exp
        # BOFT(m, b)
        self.boft_b = b
        self.boft_m = sum(int(i) for i in f"{m_exp-1:b}") + 1
        # block_num > block_size
        self.rescaled = rescaled
        self.constrain = constrain * out_dim
        self.oft_blocks = nn.Parameter(
            torch.zeros(self.boft_m, self.block_num, self.block_size, self.block_size)
        )
        if rescaled:
            self.rescale = nn.Parameter(torch.ones(out_dim))

        self.rank_dropout = rank_dropout
        self.rank_dropout_scale = rank_dropout_scale
        self.module_dropout = module_dropout

        self.multiplier = multiplier
        self.org_module = [org_module]
        self.org_forward = self.org_module[0].forward

    @property
    def I(self):
        return torch.eye(self.block_size, device=next(self.parameters()).device)

    def apply_to(self, **kwargs):
        self.org_forward = self.org_module[0].forward
        self.org_module[0].forward = self.forward

    def restore(self):
        self.org_module[0].forward = self.org_forward

    def merge_to(self, multiplier=1.0):
        weight = self.make_weight(scale=multiplier)
        self.org_module[0].weight.data.copy_(weight)

    def get_r(self):
        I = self.I
        # for Q = -Q^T
        q = self.oft_blocks - self.oft_blocks.transpose(-1, -2)
        normed_q = q
        # Diag OFT style constrain
        if self.constrain > 0:
            q_norm = torch.norm(q) + 1e-8
            if q_norm > self.constrain:
                normed_q = q * self.constrain / q_norm
        # use float() to prevent unsupported type
        r = (I + normed_q) @ (I - normed_q).float().inverse()
        return r

    def make_weight(self, scale=1, device=None):
        if self.rank_dropout and self.training:
            drop = torch.rand(self.dim, device=device) < self.rank_dropout
            drop = drop.to(self.oft_blocks.dtype)
            if self.rank_dropout_scale:
                drop /= drop.mean()
        else:
            drop = 1

        m = self.boft_m
        b = self.boft_b
        r_b = b // 2
        r = self.get_r()
        inp = self.org_module[0].weight.to(device, dtype=r.dtype)

        for i in range(m):
            bi = r[i]  # b_num, b_size, b_size
            if i == 0:
                # Apply multiplier/scale and rescale into first weight
                bi = bi * scale + (1 - scale) * self.I
                if self.rescaled:
                    bi = bi * self.rescale
            inp = rearrange(inp, "(c g k) ... -> (c k g) ...", g=2, k=2**i * r_b)
            inp = rearrange(inp, "(d b) ... -> d b ...", b=b)
            inp = torch.einsum("b i j, b j ... -> b i ...", bi, inp)
            inp = rearrange(inp, "d b ... -> (d b) ...")
            inp = rearrange(inp, "(c k g) ... -> (c g k) ...", g=2, k=2**i * r_b)

        weight = inp
        if self.rescaled:
            weight = self.rescale[:, *(None for _ in weight.shape[1:])] * weight
        return weight * drop

    @torch.no_grad()
    def apply_max_norm(self, max_norm, device=None):
        orig_norm = self.oft_blocks.to(device).norm()
        norm = torch.clamp(orig_norm, max_norm / 2)
        desired = torch.clamp(norm, max=max_norm)
        ratio = desired / norm

        scaled = ratio.item() != 1.0
        if scaled:
            self.oft_blocks *= ratio

        return scaled, orig_norm * ratio

    def forward(self, x):
        if self.module_dropout and self.training:
            if torch.rand(1) < self.module_dropout:
                return self.org_forward(x)
        scale = self.multiplier

        w = self.make_weight(scale, x.device)
        kw_dict = self.kw_dict | {"weight": w, "bias": self.org_module[0].bias}
        return self.op(x, **kw_dict)


if __name__ == "__main__":
    dim = 320
    test_layer = nn.Linear(dim, dim)
    test_layer.weight = nn.Parameter(torch.eye(dim))
    test_boft = ButterflyOFTModule("test", test_layer, lora_dim=16)
    test_boft.oft_blocks = nn.Parameter(torch.randn_like(test_boft.oft_blocks))

    print(test_boft.make_weight(1, "cpu").shape)
    print(torch.count_nonzero(test_boft.make_weight(1, "cpu")), dim * dim)



import math
from weakref import ref
from collections import OrderedDict, abc as container_abcs

import torch
import torch.nn as nn
import torch.nn.functional as F

from .base import ModuleCustomSD


class NormModule(ModuleCustomSD):
    """
    modifed from kohya-ss/sd-scripts/networks/lora:LoRAModule
    """

    def __init__(
        self,
        lora_name,
        org_module: nn.Module,
        multiplier=1.0,
        rank_dropout=0.0,
        module_dropout=0.0,
        rank_dropout_scale=False,
        **kwargs,
    ):
        """if alpha == 0 or None, alpha is rank (no scaling)."""
        super().__init__()
        self.lora_name = lora_name

        if isinstance(org_module, nn.LayerNorm):
            self.op = F.layer_norm
            self.dim = org_module.normalized_shape[0]
            self.kw_dict = {
                "normalized_shape": org_module.normalized_shape,
                "eps": org_module.eps,
            }
        elif isinstance(org_module, nn.GroupNorm):
            self.op = F.group_norm
            self.group_num = org_module.num_groups
            self.dim = org_module.num_channels
            self.kw_dict = {"num_groups": org_module.num_groups, "eps": org_module.eps}
        else:
            raise NotImplementedError
        self.w_norm = nn.Parameter(torch.zeros(self.dim))
        self.b_norm = nn.Parameter(torch.zeros(self.dim))

        self.rank_dropout = rank_dropout
        self.rank_dropout_scale = rank_dropout_scale
        self.module_dropout = module_dropout

        self.multiplier = multiplier
        self.org_module = [org_module]
        self.org_forward = self.org_module[0].forward

    def apply_to(self, **kwargs):
        self.org_forward = self.org_module[0].forward
        self.org_module[0].forward = self.forward

    def restore(self):
        self.org_module[0].forward = self.org_forward

    def merge_to(self, multiplier=1.0):
        weight, bias = self.make_weight(scale=multiplier)
        self.org_module[0].weight.data.copy_(weight)
        self.org_module[0].bias.data.copy_(bias)

    def make_weight(self, scale=1, device=None):
        org_weight = self.org_module[0].weight.to(device, dtype=self.w_norm.dtype)
        org_bias = self.org_module[0].bias.to(device, dtype=self.b_norm.dtype)
        if self.rank_dropout and self.training:
            drop = (torch.rand(self.dim, device=device) < self.rank_dropout).to(
                self.w_norm.device
            )
            if self.rank_dropout_scale:
                drop /= drop.mean()
        else:
            drop = 1
        drop = (
            torch.rand(self.dim, device=device) < self.rank_dropout
            if self.rank_dropout and self.training
            else 1
        )
        weight = self.w_norm.to(device) * drop * scale
        bias = self.b_norm.to(device) * drop * scale
        return org_weight + weight, org_bias + bias

    def forward(self, x):
        if self.module_dropout and self.training:
            if torch.rand(1) < self.module_dropout:
                return self.org_forward(x)
        scale = self.multiplier

        w, b = self.make_weight(scale, x.device)
        kw_dict = self.kw_dict | {"weight": w, "bias": b}
        return self.op(x, **kw_dict)
